class Auth {
    constructor(url, token = undefined) {
        this.username = '';
        this.url = url;
        this.llt = token; // token is an optional long-lived access token
    }

    set accessToken(t) {
        this.token = t;
    }

    get accessToken() {
        return this.token;
    }

    set expiresAt(exp) {
        this.expires = exp * 1000; // get it into milliseconds
    }

    get isExpired() {
        return Date.now() > this.expires;
    }

    login(username, password) {
        const data = { username, password };
        return new Promise((resolve, reject) => {
            fetch(this.url + '/login', 
                {
                    method: 'POST', // *GET, POST, PUT, DELETE, etc.
                    mode: 'cors', // no-cors, *cors, same-origin
                    cache: 'no-cache', // *default, no-cache, reload, force-cache, only-if-cached
                    credentials: 'include', // include, *same-origin, omit
                    headers: {
                        'Content-Type': 'application/json'
                    },
                    body: JSON.stringify(data) // body data type must match "Content-Type" header
                })
                .then(res => res.json())
                .then(res => {
                    // console.log(res)
                    if (res.username) {
                        this.accessToken = res.token;
                        this.expiresAt = res.expires;
                        resolve({status: 'success', username: res.username})
                    }
                })
                .catch(error => reject(error));
        })
    }

    logout() {
        this.accessToken = '';
        fetch(this.url + '/logout', {method: 'POST', credentials: 'include'})
    }

    refreshToken() {
        return new Promise((resolve, reject) => {
            if (this.llt) {
                // if a long-lived token was specified, use it
                this.accessToken = this.llt;
                this.expiresAt = Number.MAX_SAFE_INTEGER; // not really true but good enough as long-lived tokens have a decades-long expiration
                resolve({status: 'success', username: 'longlivedtoken'})
            } else
                fetch(this.url + '/refreshLogin', 
                    {
                        method: 'POST', // *GET, POST, PUT, DELETE, etc.
                        mode: 'cors', // no-cors, *cors, same-origin
                        cache: 'no-cache', // *default, no-cache, reload, force-cache, only-if-cached
                        credentials: 'include', // include, *same-origin, omit
                        headers: {
                            'Content-Type': 'application/json'
                        }
                    })
                .then(res => res.json())
                .then(res => {
                    console.log(res)
                    if (res.code == 'no_users') resolve({status: 'onboard'})
                    else if (res.username) {
                        this.accessToken = res.token;
                        this.expiresAt = res.expires;
                        resolve({status: 'success', username: res.username})
                    } else resolve({status: 'failure'});
                })
                .catch(error => {
                    // console.log(error)
                    reject(error)
                });
        });
    }

    createUser(username, password, isAdmin) {
        const data = { username, password, isAdmin, jwt: this.accessToken };
        return new Promise((resolve, reject) => {
            fetch(this.url + '/register', 
                {
                    method: 'POST', // *GET, POST, PUT, DELETE, etc.
                    mode: 'cors', // no-cors, *cors, same-origin
                    cache: 'no-cache', // *default, no-cache, reload, force-cache, only-if-cached
                    credentials: 'same-origin', // include, *same-origin, omit
                    headers: {
                        'Content-Type': 'application/json'
                    },
                    body: JSON.stringify(data) // body data type must match "Content-Type" header
                })
                .then(res => res.json())
                .then(res => resolve(res))
                .catch(error => reject(error));
        });
    }
    
}

module.exports = { Auth };