const EventEmitter = require('events');
const { Socket } = require('./socket.js');

const { actionsAPI, authAPI, eventsAPI, statusAPI, systemAPI } = require('./endpoints/index.js');

class WebSocketAPI extends EventEmitter {
    constructor(url, port = 2542, token = undefined) {
        super();

        this.isConnected = false;

        this._msgid = 0; // a value to incrememnt for keeping track of requests and responses
        this._waitingFor = {}; // an object for holding onto the resolve method for sent messages awaiting a response

        this.socket = new Socket(`${url}:${port}`, token);

        this.socket.on('connected', () => {
            this.isConnected = true;
            // this.emit('auth', {status: 'success', username: this.socket.username})
            // resend any messages that are still being waited on
            Object.keys(this._waitingFor).forEach(id => {
                delete this._waitingFor[id];
                // console.log(`Resending message id ${id} after connection`, this._waitingFor[id])
                // this.socket.send(this._waitingFor[id].message)
            })
            this.emit('open');
        });

        this.socket.on('message', (message) => {
            // console.log('Weboscket message received');
            switch (message.data) {
                case 'login':
                    this.emit('auth', {status: 'request'});
                    break;
                default: {
                    const { id, ...response } = JSON.parse(message.data)
                    if (id === -1) this.emit('update', response.data);
                    else if (id in this._waitingFor) {
                        this._waitingFor[id].resolve(response);
                        delete this._waitingFor[id];
                    }
                    else console.error('Got websocket response for unknown id', id, response)
                    break;
                }
            }
        });

        this.socket.on('auth', (message) => {
            switch (message.status) {
                case 'login':
                    this.emit('auth', {status: 'request'});
                    break;
                case 'success':
                    this.emit('auth', {status: 'success', username: this.socket.username})
                    break;
                case 'onboard':
                    this.emit('auth', {status: 'onboard'});
                    break;
                case 'fail':
                    this.emit('auth', {status: 'request'});
                    break;
                default: 
                    console.log('Websocket auth message default case', message)
                    break;
            }

        })

        this.socket.on('closed', () => {
            this.isConnected = false;
            this.emit('closed');
        });
        
    }

    login(username, password) {
        console.log('Logging in.')
        this.socket.login(username, password)
            .then(res => {
                this.emit('auth', {status: 'success', username: res.username});
            })
            .catch(error => {
                console.error(error)
            });
    }
    
    logout() {
        console.log('Logging out.')
        this.socket.logout();
    }

    /**
     * Send data through the OpenAVC WebSocket API
     * @param {Object} data - the data to send to the WebSocket server
     */
    async send(data) {
        console.log(data)
        // TODO: Add error handling

        const id = data._meta ? data._meta.id : ++this._msgid;
        return new Promise((resolve, reject) => {
            if (typeof data !== 'object') reject('Websocket API: Could not send. data must be an object')

            const message = {
                _meta: {
                    id,
                },
                ...data,
            };
        
            this.socket.send(message)
                .then(() => {
                    this._waitingFor[id] = {
                        resolve,
                        message
                    };
                })
                .catch (err => {
                    reject(`Websocket API: Could not send ${err}`)
                })
        })
    }

    /**
     * WebsocketAPI subscription request. Sends a subscription request to the OpenAVC server
     * @param {string} event - the OpenAVC event to subscribe to
     * @returns {boolean} True if the message was sent to the server, false otherwise
     */
    subscribe(event) {
        return new Promise((resolve, reject) => {
            this.send(eventsAPI.subscribe.buildMessage(event))
                .then((res) => {
                    resolve(eventsAPI.subscribe.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    /**
     * WebsocketAPI unsubscribe request. Sends an unsunscribe request to the OpenAVC server
     * @param {string} event - the OpenAVC event to unsubscribe from
     * @returns {boolean} True if the message was sent to the server, false otherwise
     */
    unsubscribe(event) {
        return new Promise((resolve, reject) => {
            this.send(eventsAPI.unsubscribe.buildMessage(event))
                .then((res) => {
                    resolve(eventsAPI.unsubscribe.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    /**
     * WebsocketAPI log request
     * @param {number} [maxlines=0] - The max number of lines to request
     * @returns {boolean} True if the message was sent to the server, false otherwise
     */
    logs(maxlines = 0) {
        this.send(statusAPI.logs.buildMessage(maxlines))
            .then((res) => {
                const parsed = statusAPI.logs.parseResponse(res);
                this.emit('update', {event: 'system/log', eventData: parsed });
            })
            .catch((err) => console.error('Bad log query:', err));
    }

    /**
     * WebsocketAPI context request
     */
    context({ deviceId, root }) {
        return new Promise((resolve, reject) => {
            this.send(statusAPI.context.buildMessage({deviceId, root}))
                .then((res) => {
                    resolve(statusAPI.context.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    /**
     * WebsocketAPI emit request
     * @param {Object} params - Parameter object to pass action information
     * @param {string} params.event - the openAVC event to emit
     * @param {  *   } params.data - the data to pass along to the action execution
     * @returns {boolean} True if the message was sent to the server, false otherwise
     */
    sendEvent(params) {
        return new Promise((resolve, reject) => {
            this.send(eventsAPI.emit.buildMessage(params.event, params.data))
                .then((res) => {
                    resolve(eventsAPI.emit.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    /**
     * WebsocketAPI request to call a action on the OpenAVC server
     * @param {Object} params - Parameter object to pass action information
     * @param {string} params.action - the action to execute
     * @param {  *   } params.data - the data to pass along to the action execution
     * @returns {Promise} Contains the value returned by running the action on the OpenAVC server
     */
    runAction(params) {
        return new Promise((resolve, reject) => {
            this.send(actionsAPI.action.buildMessage(params.action, params.data))
                .then((res) => {
                    resolve(actionsAPI.action.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    /**
     * WebsocketAPI request to set a variable on the OpenAVC server
     * @param {Object} params - Parameter object to pass variable information
     * @param {string} params.name - the variable name to set
     * @param {  *   } params.value - the value to set the variable to
     * @returns {Promise} Contains the value returned by running the action on the OpenAVC server
     */
    setVariable(params) {
        return new Promise((resolve, reject) => {
            this.send(actionsAPI.variable.buildMessage(params.name, params.value))
                .then((res) => {
                    resolve(actionsAPI.variable.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    getDevices() {
        this.send(statusAPI.devices.buildMessage())
            .then((res) => {
                if (res.success) {
                    const parsed = statusAPI.devices.parseResponse(res)
                    this.emit('update', {event: 'system/device', eventData: parsed });
                }
            })
            .catch((err) => console.error('Bad state query:', err));
    }

    getVariables() {
        this.send(statusAPI.variables.buildMessage())
            .then((res) => {
                if (res.success) {
                    const parsed = statusAPI.variables.parseResponse(res)
                    this.emit('update', {event: 'system/variable', eventData: parsed });
                }
            })
            .catch((err) => console.error('Bad state query:', err));
    }

    getSysInfo() {
        this.send(statusAPI.system.buildMessage())
            .then((res) => {
                const parsed = statusAPI.system.parseResponse(res)
                this.emit('update', {event: 'system/system', eventData: parsed });
            })
            .catch((err) => console.error('Bad system query:', err));
    }

    restart() {
        return new Promise((resolve, reject) => {
            this.send(systemAPI.restart.buildMessage(name))
                .then((res) => {
                    resolve(systemAPI.restart.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    reloadConstants() {
        return new Promise((resolve, reject) => {
            this.send(systemAPI.reloadConstants.buildMessage())
                .then((res) => {
                    resolve(systemAPI.reloadConstants.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    runScript(name) {
        return new Promise((resolve, reject) => {
            this.send(systemAPI.script.buildMessage(name))
                .then((res) => {
                    resolve(systemAPI.script.parseResponse(res))
                })
                .catch((err) => reject(err));
        });
    }

    createToken(name) {
        return new Promise((resolve, reject) => {
            this.send(authAPI.createToken.buildMessage(name))
                .then((res) => {
                    resolve(authAPI.createToken.parseResponse(res));
                })
                .catch((err) => reject(err));
        });
    }

    revokeToken(name) {
        return new Promise((resolve, reject) => {
            this.send(authAPI.revokeToken.buildMessage(name))
                .then((res) => {
                    resolve(authAPI.revokeToken.parseResponse(res));
                })
                .catch((err) => reject(err));
        });
    }

    createUser(username, password, isAdmin) {
        return this.socket.createUser(username, password, isAdmin)
    }
}

module.exports = { WebSocketAPI };

