const EventEmitter = require('events');
const { Auth }   = require('./auth.js');

const RECONNECT_INTERVAL = 5000;
const PING_TIMEOUT_INTERVAL = 10000 + 5000; // heartbeat sent by server every 10 seconds plus a buffer of 5 seconds before assuming the connection died


class Socket extends EventEmitter{
    constructor(url, token = undefined) {
        super();
        this.url = url;
        this.username = '';
        this.socket = null;
        this.auth = new Auth(this.httpUrl, token);

        this.refreshToken();

        this.pingTimeout = null;
        this.reconnectTimeout = null;
    }

    get wsUrl() {
        return `ws://${this.url}`
    }

    get httpUrl() {
        return `http://${this.url}`
    }

    get isExpired() {
        return this.auth.isExpired;
    }

    connect() {
        if (this.socket && (this.socket.readyState == 1 || this.socket.readyState == 0)) return; // connect{ed,ing}
        
        const opened = () => {
            console.log('Websocket connected');
            this.heartbeat();
            this.emit('connected');
        };

        // eslint-disable-next-line no-unused-vars
        const closed = (event) => {
            this.close();
            console.log('Websocket closed');
            this.emit('closed');
            this.reconnectTimeout = setTimeout(() => {
                this.connect();
            }, RECONNECT_INTERVAL);
        };

        // const errored = (error) => {
        //     console.warn('Websocket error received', error);
        // };

        const newMessage = (message) => {
            switch (message.data) {
                case 'ping':
                    if (this.socket) {
                        this.socket.send('pong');
                        this.heartbeat();
                    }
                    break;
                // case 'onboard':
                //     this.emit('auth', { status: 'onboard' });
                //     break;
                // case 'login':
                //     this.emit('auth', { status: 'request' });
                //     break;
                case 'refresh':
                    this.refreshToken();
                    break;
                default: 
                    this.emit('message', message);
                    break;
            }

        }

        console.log('Attempting Websocket connection');
        try {
            this.socket = new WebSocket(this.wsUrl + `?token=${this.auth.token}`);
    
            this.socket.onopen = opened;
            this.socket.onmessage = newMessage;
            // this.socket.onerror = errored;
            this.socket.onclose = closed;
        } catch (err) {
            console.error('Websocket creation error', err)
        }

    }

    close() {
        if (this.reconnectTimeout) clearTimeout(this.reconnectTimeout);
        if (this.pingTimeout) clearTimeout(this.pingTimeout);
        if (this.socket) {
            this.socket.close();
            delete this.socket;
        }
    }

    heartbeat() {
        clearTimeout(this.pingTimeout);

        // Use `WebSocket#terminate()`, which immediately destroys the connection,
        // instead of `WebSocket#close()`, which waits for the close timer.
        // Delay should be equal to the interval at which your server
        // sends out pings plus a conservative assumption of the latency.
        this.pingTimeout = setTimeout(() => {
            console.warn('WebSocket connection timed out. Closing connection.');
            this.close(); // this.socket.onclose will try to reconnect periodically
        }, PING_TIMEOUT_INTERVAL);
    }

    login(username, password) {
        return new Promise((resolve, reject) => {
            this.auth.login(username, password)
                .then(res => {
                    this.username = res.username
                    this.emit('auth', { status: 'success' })
                    this.connect();
                    resolve(res)
                })
                .catch(error => reject(error));
        });
    }

    logout() {
        this.auth.logout();
        this.autoReconnect = false;
        this.close();
        this.emit('auth', { status: 'login' })
    }

    async refreshToken() {
        await this.auth.refreshToken()
            .then(res => {
                if (res.status == 'success') {
                    this.username = res.username
                    this.emit('auth', { status: 'success' })
                    this.connect();
                } else if (res.status =='onboard') this.emit('auth', { status: 'onboard' })
                else this.emit('auth', { status: 'fail' })
            })
            .catch((error) => {
                console.error(error);
                setTimeout(() => {
                    this.refreshToken();
                }, 5000)
            });
    }

    /**
     * Send data through the OpenAVC WebSocket API
     * @param {Object} message - the message to send to the WebSocket server
     */
    async send(message) {
        // TODO: Add error handling

        if (this.auth.isExpired) await this.refreshToken();

        return new Promise((resolve, reject) => {
            if (typeof message !== 'object') reject('Websocket API: Could not send. message must be an object')
            if (message._meta) message._meta.token = this.auth.token;
            else message._meta = { token: this.auth.token};
        
            this._waitForReadyState(JSON.stringify(message))
                .then(() => resolve())
                .catch (err => {
                    reject(`Websocket API: Could not send ${err}`)
                })
        })
    }

    _waitForReadyState(data, depth = 1) {
        return new Promise((resolve, reject) => {
            if (depth > 50) return reject('Maximum _waitForReadyState depth reached');
            // console.log('socket.send attempt ' + depth)
            if (this.socket && this.socket.readyState == 1) {
                resolve(this.socket.send(data));
            } else {
                const timeout = Math.min(250*depth, 5000);
                setTimeout(() => {
                    this._waitForReadyState(data, ++depth)
                    .then(res => resolve(res))
                    .catch(err => reject(err))
                }, timeout);
            }
        })
    }

    createUser(username, password, isAdmin) {
        return this.auth.createUser(username, password, isAdmin)
    }
}

module.exports = {
    Socket
}

